!> Module containing classes representing neurons
!! in neural networks.
!!
!! It uses 'time_measurement_m' and 'normal_m' modules.
!!
!! @author Martin Beseda
!! @author Martin Mrovec
!! @date 2017
module neuron_m
    use time_measurement_m
    use normal_m
    use data_kinds_4neuro_m
    use abstract_base_m

    implicit none

    public

    !----------------!------------------------------------------------------------------------------
    ! class neuron_t !
    !----------------!

    !> Abstract class representing a general neuron
    type, abstract, extends(abstract_base_t) :: neuron_t
        !TODO properties are public because of problems with reading from binary file - find another way

        real(kind=real_4neuro)       :: potential  !< Neuron inner potential
        real(kind=real_4neuro)       :: state      !< State of the neuron (0/1 for binary n. etc.)
        integer(kind=integer_4neuro) :: id         !< Neuron ID

    contains
        !> Initalizes the common neuron_t classes components 
        !! ('potential' to 0, 'state' to 0)
        procedure, private                         :: init_components    => neuron_init_components_impl

        !> Adds the input signal value to the current potential
        !! @param[in] input_signal Input value
        procedure                                  :: adjust_potential   => adjust_potential_impl

        !> Performs the activation function and stores the result into the 'state' component
        procedure(activate_int), private, deferred :: activate

        !> Getter to the 'potential' component
        !! @return Value of the neuron's potential (real number)
        procedure                                  :: get_potential      => get_potential_impl

        !> Getter to the 'state' component
        !! @return Value of the current neuron's state
        procedure                                  :: get_state          => get_state_impl

        !> Getter to the 'id' component
        !! @return Neuron ID
        procedure                                  :: get_id             => get_id_impl

        !> Setter to the 'potential' component
        procedure                                  :: set_potential      => set_potential_impl

        !> Setter to the 'state' component
        procedure                                  :: set_state          => set_state_impl

        !> Setter to the 'id' component
        procedure                                  :: set_id             => set_id_impl

        !> Implementation of write function enabling the storage of
        !! allocatable arrays.
        procedure                                  :: write_sample       => write_neuron_sample

        !> Implementation of read function enabling to read
        !! the whole instance of net_t stored as binary file.
        procedure                                  :: read_sample        => read_neuron_sample

        generic                                    :: write(unformatted) => write_sample
        generic                                    :: read(unformatted)  => read_sample
    end type neuron_t

    abstract interface
        subroutine activate_int(this)
            !> Interface for the deferred (i.e. abstract) method 'activate' of  the clas neuron_t

            import neuron_t

            class(neuron_t), intent(inout) :: this
        end subroutine activate_int
    end interface

    !---------------------!------------------------------------------------------------------------
    ! class mock_neuron_t !
    !---------------------!

    !> Mock neuron_t class 
    type, extends(neuron_t) :: mock_neuron_t
    contains

        !> Mock method, does nothing - it's only purpose is to override the deffered one
        procedure :: activate      => mock_activate_impl

        !> Getter to the 'potential' component
        !! @return Number 5.0
        procedure :: get_potential => mock_get_potential_impl

        !> Getter to the 'state' component
        !! @return Number 15.0
        procedure :: get_state     => mock_get_state_impl
    end type mock_neuron_t

    interface mock_neuron_t
        !> Constructor of mock_neuron_t class
        !! @return Returns an instance of the class mock_neuron_t
        module procedure :: new_mock_neuron
    end interface mock_neuron_t

    !-----------------------!----------------------------------------------------------------------
    ! class binary_neuron_t !
    !-----------------------!

    !> Binary neuron class - uses unit-step as the activation function
    type, extends(neuron_t) :: binary_neuron_t
        private

        real(kind=real_4neuro) :: threshold  !< When neuron potential exceeds this value, neuron becomes excited

    contains

        !> Activation function - transforms potential into the output state value
        !! AND assigns it into the 'state' component
        !!
        !! Unit-step function - (returns 1 for potential > threshold, otherwise returns 0)
        procedure, private :: activate => unit_step_activate_impl

    end type binary_neuron_t

    interface binary_neuron_t
        !> Non-parametric constructor of binary_neuron_t class (threshold
        !! will be initialized by a random number from Gaussian distribution)
        !! @return An instance of the class binary_neuron_t 
        module procedure :: new_binary_neuron

        !> Parametric constructor of binary_neuron_t class
        !! @param[in] threshold Threshold for the unit-step activation function
        !! @return An instance of the class binary_neuron_t 
        module procedure :: new_binary_neuron_1
    end interface binary_neuron_t

    !-----------------------!----------------------------------------------------------------------
    ! class linear_neuron_t !
    !-----------------------!

    !> Linear neuron class - uses activation function in the form f(x)=a*x + b,
    !! 'x' being the neuron's potential
    type, extends(neuron_t) :: linear_neuron_t
        private

        !! Coefficients for the linear activation function in format 'f(x)=a*x + b'
        real(kind=real_4neuro) :: a_coef  !< The coefficient 'a' in the activation function f(x)=a*x + b
        real(kind=real_4neuro) :: b_coef  !< The coefficient 'b' in the activation function f(x)=a*x + b

        contains

            !> Activation function - f(x)=a*x + b
            procedure, private :: activate => identity_potential_activate_impl
    end type linear_neuron_t

    interface linear_neuron_t
        !> Non-parametric constructor of linear_neuron_t class - a coef. is 1, b coef. is 0
        !! @return An instance of the class linear_neuron_t
        module procedure :: new_linear_neuron

        !> Constructor of linear_neuron_t class
        !! @param[in] a_coef a coef. of the linear activation function
        !! @param[in] b_coef b coef. of the linear activation fucntion
        !! @return An instance of the class linear_neuron_t
        module procedure :: new_linear_neuron_2
    end interface linear_neuron_t

    !-------------------------!--------------------------------------------------------------------
    ! class logistic_neuron_t !
    !-------------------------!

    !> Logistic neuron class - uses generalised logistic function as an activation function
    !! in the form f(x) = (1 + e^(-x))^(-alpha),
    !! 'x' being the neuron potential here
    type, extends(neuron_t) :: logistic_neuron_t
        private 

        real(kind=real_4neuro) :: alpha_coef  !< The alpha coefficient used in the activation function

        contains

            !> Activation function - generalised logistic f.
            procedure, private :: activate => logistic_activate_impl
    end type logistic_neuron_t

    interface logistic_neuron_t
        !> Non-parametric constructor of logistic_neuron_t class
        !! Alpha coefficient is set to 1
        !! @return An instance of the class logistic_neuron_t
        module procedure :: new_logistic_neuron

        !> Constructor of the logistic_neuron_t class
        !! @param[in] alpha_coef Alpha coefficient in the logistic activation function
        !! @return An instance of the class logistic_neuron_t
        module procedure :: new_logistic_neuron_1
    end interface logistic_neuron_t

    !---------------------!------------------------------------------------------------------------
    ! class tanh_neuron_t !
    !---------------------!

    !> Hyperbolic tangent neuron class - uses f(x) = (e^x - e^(-x))/(e^x + e^(-x)) as 
    !! an activation function, 'x' being the neuron potential.
    type, extends(neuron_t) :: tanh_neuron_t
        contains

            !> Activation function - hyperbolic tangent
            procedure, private :: activate => hyperbolic_tangent_activate_impl
    end type tanh_neuron_t

    interface tanh_neuron_t
        !> Constructor for an instance of the class tanh_neuron_t
        module procedure :: new_tanh_neuron
    end interface tanh_neuron_t

    !> Arcus tangents neuron class - uses f(x)=tan^(-1)(x) as an
    !! activation function, 'x' being the neuron potential
    type, extends(neuron_t) :: arctan_neuron_t
        contains

            !> Activation function - arcus tangens
            procedure, private :: activate => arcus_tangens_activate_impl
    end type arctan_neuron_t

    interface arctan_neuron_t
        !> Constructor for an instance of the class arctan_neuron_t
        module procedure :: new_arctan_neuron
    end interface arctan_neuron_t

    contains
        !------------------------!---------------------------------------------------------------------
        ! Method implementations !
        !------------------------!

        !----------------!
        ! class neuron_t !
        !----------------!

        !-------------------!--------------------------------------------------------------------------
        ! Getters & Setters !
        !-------------------!

        !> Getter to the 'potential' component
        !! @return Value of the neuron's potential (real number)
        function get_potential_impl(this) result(potential)
            class(neuron_t), intent(in) :: this
            real(kind=real_4neuro)      :: potential
#ifdef TIME_PROFILING
            real                        :: start_time
            call time_profiling_start(start_time)
#endif 
            potential = this%potential

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'get_potential_impl')
#endif
        end function get_potential_impl

        !> Getter to the 'state' component
        !! @return Value of the current neuron's state
        function get_state_impl(this) result(state)
            class(neuron_t), intent(in) :: this
            real(kind=real_4neuro)      :: state
#ifdef TIME_PROFILING
            real                        :: start_time
            call time_profiling_start(start_time)
#endif
            state = this%state
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'get_state_impl')
#endif
        end function get_state_impl

        !> Getter to the 'id' component
        !! @return Neuron ID
        function get_id_impl(this) result(id)
            class(neuron_t), intent(in)  :: this
            integer(kind=integer_4neuro) :: id
#ifdef TIME_PROFILING
            real                         :: start_time
            call time_profiling_start(start_time)
#endif
            id = this%id
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'get_id_impl')
#endif
        end function get_id_impl

        !> Setter to the 'potential' component
        !! @param[in] potential The value of potential to be set
        subroutine set_potential_impl(this, potential)
            class(neuron_t), intent(inout)     :: this
            real(kind=real_4neuro), intent(in) :: potential
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            this%potential = potential
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'set_potential_impl')
#endif
        end subroutine set_potential_impl

        !> Setter to the 'state' component
        !! @param[in] state The state to be set
        subroutine set_state_impl(this, state)
            class(neuron_t), intent(inout)     :: this
            real(kind=real_4neuro), intent(in) :: state
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            this%state = state
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'set_state_impl')
#endif
        end subroutine set_state_impl

        !> Setter to the 'id' component
        !! @param[in] id The ID of neuron to be set
        subroutine set_id_impl(this, id)
            class(neuron_t), intent(inout)           :: this
            integer(kind=integer_4neuro), intent(in) :: id
#ifdef TIME_PROFILING
            real                                     :: start_time
            call time_profiling_start(start_time)
#endif
            this%id = id
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'set_id_impl')
#endif
        end subroutine set_id_impl


        !----------------!-----------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Initalizes the common neuron_t classes components 
        !! ('potential' to 0, 'state' to 0)
        subroutine neuron_init_components_impl(this)
            class(neuron_t), intent(inout) :: this
            integer(kind=integer_4neuro)   :: values(8)  !< values(8) used as a seed
#ifdef TIME_PROFILING
            real                           :: start_time
            call time_profiling_start(start_time)
#endif 
            this%potential = 0
            this%state     = 0  !TODO maybe init with values from normal distribution
            this%id        = 1  !TODO modify - should be index in the whole set of neurons

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'neuron_init_components_impl')
#endif

        end subroutine neuron_init_components_impl

        !> Adds the input signal value to the current potential
        !! @param[in] input_signal Input value
        subroutine adjust_potential_impl(this, input_signal)
            class(neuron_t), intent(inout)     :: this
            real(kind=real_4neuro), intent(in) :: input_signal
#ifdef TIME_PROFILING
            real                           :: start_time
            call time_profiling_start(start_time)
#endif
            this%potential = this%potential + input_signal
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'adjust_potential_impl')
#endif
        end subroutine adjust_potential_impl

        !> Implementation of write function enabling the storage of
        !! allocatable arrays.
        !! @param[in] unit Unit (handler) where the file is written
        !! @param[out] iostat Diagnostic value "returned" by the subroutine
        !! @param[inout] iomsg Explaining note about error
        subroutine write_neuron_sample(this, unit, iostat, iomsg)
            class(neuron_t), intent(in) :: this
            integer, intent(in)         :: unit
            integer, intent(out)        :: iostat
            character(*), intent(inout) :: iomsg
#ifdef TIME_PROFILING
            real                        :: start_time
            call time_profiling_start(start_time)
#endif
            write(unit, iostat=iostat, iomsg=iomsg) this%potential, &
                                                    this%state, &
                                                    this%id
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'write_neuron_sample')
#endif
        end subroutine write_neuron_sample

        ! Unformatted reading for the sample derived type
        subroutine read_neuron_sample(this, unit, iostat, iomsg)
            class(neuron_t), intent(inout) :: this
            integer, intent(in)            :: unit
            integer, intent(out)           :: iostat
            character(*), intent(inout)    :: iomsg
#ifdef TIME_PROFILING
            real                           :: start_time
            call time_profiling_start(start_time)
#endif
            ! And then finally the reading.
            read(unit, iostat=iostat, iomsg=iomsg) this%potential, &
                                                   this%state, &
                                                   this%id
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'read_neuron_sample')
#endif
    end subroutine read_neuron_sample

        !---------------------!
        ! class mock_neuron_t !
        !---------------------!

        !----------------!----------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Mock method, does nothing - it's only purpose is to override the deffered one
        subroutine mock_activate_impl(this)
            class(mock_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real :: start_time
            call time_profiling_start(start_time)
#endif

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'mock_activate_impl')
#endif
        end subroutine

        !--------------!------------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Constructor of mock_neuron_t class
        !! @return Returns an instance of the class mock_neuron_t
        function new_mock_neuron() result(new_obj)
            type(mock_neuron_t) :: new_obj
#ifdef TIME_PROFILING
            real                          :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_mock_neuron')
#endif
        end function new_mock_neuron

        !---------!-----------------------------------------------------------------------------------
        ! Getters !
        !---------!

        !> Getter to the 'potential' component
        !! @return Number 5.0
        function mock_get_potential_impl(this) result(return_value)
            class(mock_neuron_t), intent(in) :: this
            real(kind=real_4neuro)           :: return_value
#ifdef TIME_PROFILING
            real                             :: start_time
            call time_profiling_start(start_time)
#endif
            return_value = this%potential

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'mock_get_potential_impl')
#endif
        end function mock_get_potential_impl

        !> Getter to the 'state' component
        !! @return Number 15.0
        function mock_get_state_impl(this) result(return_value)
            class(mock_neuron_t), intent(in) :: this
            real(kind=real_4neuro)           :: return_value
#ifdef TIME_PROFILING
            real                             :: start_time
            call time_profiling_start(start_time)
#endif
            return_value = 15.0
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'mock_get_state_impl')
#endif
        end function mock_get_state_impl

        !-----------------------!
        ! class binary_neuron_t !
        !-----------------------!

        !--------------!-----------------------------------------------------------------------------
        ! Constructors !
        !--------------!
        
        !> Non-parametric constructor of binary_neuron_t class (threshold
        !! will be initialized by a random number from Gaussian distribution)
        !! @return An instance of the class binary_neuron_t 
        function new_binary_neuron() result(new_obj)
            type(binary_neuron_t) :: new_obj
            integer(kind=4)       :: values(8)
#ifdef TIME_PROFILING
            real                  :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()

            ! Get current time and use milliseconds (values(8))
            ! as a seed
            call date_and_time(values=values)

            ! Generate random number from Gaussian distribution
            ! and use it to initialize the component 'threshold'
            new_obj%threshold = r4_normal_01(values(8))
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_binary_neuron')
#endif
        end function new_binary_neuron

        !> Constructor of binary_neuron_t class
        !! @param[in] threshold Threshold used by unit-step activation function
        !! @return An instance of the class binary_neuron_t 
        function new_binary_neuron_1(threshold) result(new_obj)
            real(kind=real_4neuro), intent(in) :: threshold
            type(binary_neuron_t)              :: new_obj
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()
            new_obj%threshold = threshold
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_binary_neuron_1')
#endif
        end function new_binary_neuron_1

        !----------------!---------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Activation function - transforms potential into the output state value
        !! AND assigns it into the 'state' component
        !!
        !! Unit-step function - (returns 1 for potential > threshold, otherwise returns 0)       
        subroutine unit_step_activate_impl(this)
            class(binary_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                                  :: start_time
            call time_profiling_start(start_time)
#endif
            
            if(this%potential > this%threshold) then
                this%state = 1
            else
                this%state = 0
            end if

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'unit_step_activate_impl')
#endif
        end subroutine unit_step_activate_impl

        !-----------------------!
        ! class linear_neuron_t !
        !-----------------------!

        !--------------!-----------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Non-parametric constructor of linear_neuron_t class - a coef. is 1, b coef. is 0
        !! @return An instance of the class linear_neuron_t
        function new_linear_neuron() result(new_obj)
            type(linear_neuron_t) :: new_obj
#ifdef TIME_PROFILING
            real                  :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()

            new_obj%a_coef = 1
            new_obj%b_coef = 0

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_linear_neuron')
#endif
        end function new_linear_neuron

        !> Constructor of linear_neuron_t class
        !! @param[in] a_coef a coef. of the linear activation function
        !! @param[in] b_coef b coef. of the linear activation fucntion
        !! @return An instance of the class linear_neuron_t
        function new_linear_neuron_2(a_coef, b_coef) result(new_obj)
            real(kind=real_4neuro), intent(in) :: a_coef
            real(kind=real_4neuro), intent(in) :: b_coef

            type(linear_neuron_t)              :: new_obj
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()

            new_obj%a_coef = a_coef
            new_obj%b_coef = b_coef

#ifdef TIME_PROFILING
             call time_profiling_stop(start_time, 'new_linear_neuron_2')
#endif   
        end function new_linear_neuron_2

        !----------------!---------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Activation function - just returns the neuron's potential value
        subroutine identity_potential_activate_impl(this)
            class(linear_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                                  :: start_time
            call time_profiling_start(start_time)
#endif

            this%state = this%potential

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'identity_potential_activate_impl')
#endif
        end subroutine identity_potential_activate_impl

        !-------------------------!
        ! class logistic_neuron_t !
        !-------------------------!

        !--------------!-----------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Non-parametric constructor of logistic_neuron_t class
        !! Alpha coefficient is set to 1
        !! @return An instance of the class logistic_neuron_t
        function new_logistic_neuron() result(new_obj)
            type(logistic_neuron_t) :: new_obj
#ifdef TIME_PROFILING
            real                    :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()
            new_obj%alpha_coef = 1

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_logistic_neuron')
#endif
        end function new_logistic_neuron

        !> Constructor of the logistic_neuron_t class
        !! @param[in] alpha_coef Alpha coefficient in the logistic activation function
        !! @return An instance of the class logistic_neuron_t
        function new_logistic_neuron_1(alpha_coef) result(new_obj)
            real(kind=real_4neuro), intent(in) :: alpha_coef
            type(logistic_neuron_t)            :: new_obj
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()
            new_obj%alpha_coef = alpha_coef

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_logistic_neuron_1')
#endif
        end function new_logistic_neuron_1

        !----------------!---------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Activation function - logistic function
        subroutine logistic_activate_impl(this)
            class(logistic_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                                    :: start_time
            call time_profiling_start(start_time)
#endif
            this%state = (1.0 + exp(-1*this%potential))**(-1*this%alpha_coef)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'logistic_activate_impl')
#endif
        end subroutine logistic_activate_impl

        !---------------------!
        ! class tanh_neuron_t !
        !---------------------!

        !--------------!------------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Constructor for an instance of the class tanh_neuron_t
        function new_tanh_neuron() result(new_obj)
            type(tanh_neuron_t) :: new_obj
#ifdef TIME_PROFILING
            real                :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_tanh_neuron')
#endif
        end function new_tanh_neuron

        !----------------!----------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Activation function - hyperbolic tangent
        subroutine hyperbolic_tangent_activate_impl(this)
            class(tanh_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                                :: start_time
            call time_profiling_start(start_time)
#endif

#if PREC == DOUBLE
            this%state = dtanh(this%potential)
#else
            this%state = tanh(this%potential)
#endif

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'hyperbolic_tangent_activate_impl')
#endif
        end subroutine hyperbolic_tangent_activate_impl

        !-----------------------!
        ! class arctan_neuron_t !
        !-----------------------!

        !--------------!------------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Constructor for an instance of the class arctan_neuron_t
        function new_arctan_neuron() result(new_obj)
            type(arctan_neuron_t) :: new_obj
#ifdef TIME_PROFILING
            real                  :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components()

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_arctan_neuron')
#endif
        end function new_arctan_neuron

        !----------------!----------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Activation function - arcus tangens
        subroutine arcus_tangens_activate_impl(this)
            class(arctan_neuron_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                                  :: start_time
            call time_profiling_start(start_time)
#endif

#if PREC == DOUBLE
            this%state = datan(this%potential)
#else
            this%state = atan(this%potential)
#endif

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'arcus_tangens_activate_impl')
#endif
        end subroutine arcus_tangens_activate_impl

end module neuron_m
